# -*- coding: utf-8 -*-
import json

from mock import MagicMock, patch
from twisted.internet.defer import succeed, fail
from twisted.python.failure import Failure
from twisted.web import server
from twisted.web.error import Error, UnsupportedMethod
from twisted.web.server import Request
from twisted.trial import unittest

from scrapyrt.resources import ServiceResource


class TestServiceResource(unittest.TestCase):

    def setUp(self):
        self.resource = ServiceResource()
        self.request = MagicMock(spec=Request)
        self.request.code = 200

        def set_code(code):
            self.request.code = code

        self.request.setResponseCode.side_effect = set_code


@patch('scrapyrt.resources.log.err')
@patch('twisted.web.resource.Resource.render')
class TestRender(TestServiceResource):

    def setUp(self):
        super(TestRender, self).setUp()
        self.request_write_values = []

        def request_write(value):
            self.request_write_values.append(value)

        self.request.write.side_effect = request_write

    def test_render_ok(self, render_mock, log_err_mock):
        render_mock.return_value = {'status': 'ok'}
        result = self.resource.render(self.request)
        obj = json.loads(result.decode("utf8"))
        self.assertIn('status', obj)
        self.assertEqual(obj['status'], 'ok')
        self.assertFalse(log_err_mock.called)

    def test_render_exception(self, render_mock, log_err_mock):
        exc = Exception('boom')
        render_mock.side_effect = exc
        result = self.resource.render(self.request)
        obj = json.loads(result.decode("utf8"))
        self.assertTrue(log_err_mock.called)
        self.assertEqual(obj['status'], 'error')
        self.assertEqual(obj['message'], str(exc))
        self.assertEqual(obj['code'], 500)

    def test_render_deferred_succeed(self, render_mock, log_err_mock):
        render_mock.return_value = succeed({'status': 'ok'})
        result = self.resource.render(self.request)
        self.assertEqual(result, server.NOT_DONE_YET)
        self.assertTrue(self.request.write.called)
        self.assertTrue(self.request.finish.called)
        self.assertEqual(len(self.request_write_values), 1)
        obj = json.loads(self.request_write_values[0].decode("utf8"))
        self.assertIn('status', obj)
        self.assertEqual(obj['status'], 'ok')
        self.assertFalse(log_err_mock.called)

    def test_render_deferred_fail(self, render_mock, log_err_mock):
        exc = Exception('boom')
        render_mock.return_value = fail(exc)
        result = self.resource.render(self.request)
        self.assertEqual(result, server.NOT_DONE_YET)
        self.assertTrue(self.request.write.called)
        self.assertTrue(self.request.finish.called)
        self.assertEqual(len(self.request_write_values), 1)
        obj = json.loads(self.request_write_values[0].decode("utf8"))
        self.assertEqual(obj['status'], 'error')
        self.assertEqual(obj['message'], str(exc))
        self.assertEqual(obj['code'], 500)
        self.assertTrue(log_err_mock.called)


@patch('twisted.python.log.msg')
class TestHandleErrors(TestServiceResource):

    def setUp(self):
        super(TestHandleErrors, self).setUp()

    def _assert_log_err_called(self, log_msg_mock, failure):
        log_msg_mock.call_count = 1
        _, kwargs = log_msg_mock.call_args
        self.assertEqual(kwargs['isError'], 1)
        self.assertEqual(kwargs['system'], 'scrapyrt')
        if isinstance(failure, Failure):
            self.assertEqual(kwargs['failure'], failure)
        else:
            self.assertEqual(kwargs['failure'].value, failure)

    def test_exception(self, log_msg_mock):
        try:
            raise Exception('blah')
        except Exception as e:
            result = self.resource.handle_error(e, self.request)
            exc = e
        self.assertEqual(self.request.code, 500)
        self.assertEqual(result['message'], str(exc))
        self._assert_log_err_called(log_msg_mock, exc)

    def test_failure(self, log_msg_mock):
        exc = Exception('blah')
        failure = Failure(exc)
        result = self.resource.handle_error(failure, self.request)
        self.assertEqual(self.request.code, 500)
        self.assertEqual(result['message'], str(exc))
        self._assert_log_err_called(log_msg_mock, failure)

    def test_error_400(self, log_msg_mock):
        exc = Error('400', 'blah_400')
        result = self.resource.handle_error(exc, self.request)
        self.assertEqual(self.request.code, 400)
        self.assertFalse(log_msg_mock.called)
        self.assertEqual(result['message'], exc.message)

    def test_error_403(self, log_msg_mock):
        exc = Error('403', 'blah_403')
        result = self.resource.handle_error(exc, self.request)
        self.assertEqual(self.request.code, 403)
        self.assertFalse(log_msg_mock.called)
        self.assertEqual(result['message'], exc.message)

    def test_error_not_supported_method(self, log_msg_mock):
        exc = UnsupportedMethod(['GET'])
        result = self.resource.handle_error(exc, self.request)
        self.assertEqual(self.request.code, 405)
        self.assertFalse(log_msg_mock.called)
        self.assertIn('GET', result['message'])


class TestFormatErrorResponse(TestServiceResource):

    def test_format_error_response(self):
        code = 400
        self.request.code = code
        exc = Error(str(code), 'blah')
        response = self.resource.format_error_response(exc, self.request)
        self.assertEqual(response['status'], 'error')
        self.assertEqual(response['message'], exc.message)
        self.assertEqual(response['code'], code)


class TestRenderObject(TestServiceResource):

    def setUp(self):
        super(TestRenderObject, self).setUp()
        self.obj = {'status': 'ok', 'key': 'value'}
        self.headers = []
        self.request = MagicMock(spec=Request)

        def add_header(name, value):
            self.headers.append((name, value))

        self.request.setHeader.side_effect = add_header

    def test_render_object(self):
        result = self.resource.render_object(self.obj, self.request)
        set_header_mock = self.request.setHeader
        self.assertEqual(set_header_mock.call_count, 5)
        set_header_mock.assert_any_call('Content-Type', 'application/json')
        set_header_mock.assert_any_call('Access-Control-Allow-Origin', '*')
        set_header_mock.assert_any_call('Access-Control-Allow-Headers',
                                        'X-Requested-With')
        set_header_mock.assert_any_call('Content-Length', str(len(result)))
        # request.setHeader('Access-Control-Allow-Methods',
        #                   ', '.join(getattr(self, 'allowedMethods', [])))
        headers = dict(self.headers)
        self.assertIn('Access-Control-Allow-Methods', headers)
        self.assertEqual(headers['Access-Control-Allow-Methods'], '')
        for key, value in self.obj.items():
            self.assertIn(key.encode("utf8"), result)
            self.assertIn(value.encode("utf8"), result)

    def _test_access_control_allow_methods_header(self):
        headers = dict(self.headers)
        self.assertIn('Access-Control-Allow-Methods', headers)
        access_control_allow_methods = headers['Access-Control-Allow-Methods']
        self.assertEqual(
            self.resource.allowedMethods,
            [s.strip() for s in access_control_allow_methods.split(',')]
        )

    def test_access_control_allow_methods_header_get(self):
        self.resource.allowedMethods = ['GET']
        self.resource.render_object(self.obj, self.request)
        self._test_access_control_allow_methods_header()

    def test_access_control_allow_methods_header_get_post(self):
        self.resource.allowedMethods = ['GET', 'POST']
        self.resource.render_object(self.obj, self.request)
        self._test_access_control_allow_methods_header()



# -*- coding: utf-8 -*-
"""Default scrapyrt settings."""

# Project settings module - found at server initialization
PROJECT_SETTINGS = None

# Path to server log file
LOG_FILE = None

# Spider logs will be kept in file with name set to timestamp in following
# format
SPIDER_LOG_FILE_TIMEFORMAT = '%Y-%m-%dT%H%M%S.%f'

# Path to spiders log directory
LOG_DIR = 'logs'

LOG_ENCODING = 'utf-8'

# Root server resource, should inherit from scrapyrt.resources.RealtimeAPI
SERVICE_ROOT = 'scrapyrt.resources.RealtimeApi'

# Resources list
RESOURCES = {
    'crawl.json': 'scrapyrt.resources.CrawlResource',
}

CRAWL_MANAGER = 'scrapyrt.core.CrawlManager'

# Limit spider run time
TIMEOUT_LIMIT = 1000
# disable in production
DEBUG = True



# -*- coding: utf-8 -*-


class ScrapyrtDeprecationWarning(Warning):
    """Warning category for deprecated features, since the default
    DeprecationWarning is silenced on Python 2.7+
    """
    pass



from mock import patch

from scrapyrt.decorators import deprecated
from scrapyrt.exceptions import ScrapyrtDeprecationWarning


class TestDecorators(object):
    def test_deprecated(self):
        @deprecated
        def test_func(*args, **kwargs):
            return (args, kwargs)

        with patch('warnings.warn') as w:
            args, kwargs = test_func('foo', 'bar', keyword='blue')

        msg = 'Call to deprecated function test_func.'
        w.assert_called_once_with(msg,
                                  category=ScrapyrtDeprecationWarning,
                                  stacklevel=2)
        assert args == ('foo', 'bar')
        assert kwargs['keyword'] == 'blue'



# -*- coding: utf-8 -*-
import os

import shutil
from scrapy.settings import Settings

from . import TESTS_PATH, SAMPLE_DATA

LOCALHOST = 'localhost'


def get_testenv():
    env = os.environ.copy()
    env['PYTHONPATH'] = os.path.realpath(os.path.join(TESTS_PATH, '..'))
    return env


def get_settings():
    """Settings with all extensions disabled."""
    return Settings({
        'EXTENSIONS': {
            'scrapy.extensions.throttle.AutoThrottle': None,
            'scrapy.extensions.feedexport.FeedExporter': None,
            'scrapy.extensions.logstats.LogStats': None,
            'scrapy.extensions.closespider.CloseSpider': None,
            'scrapy.extensions.corestats.CoreStats': None,
            'scrapy.extensions.memusage.MemoryUsage': None,
            'scrapy.extensions.memdebug.MemoryDebugger': None,
            'scrapy.extensions.spiderstate.SpiderState': None,
            'scrapy.extensions.telnet.TelnetConsole': None,
        }
    })


def generate_project(directory, site=None):
    source = os.path.join(SAMPLE_DATA, 'testproject')
    shutil.copytree(
        source, directory, ignore=shutil.ignore_patterns('*.pyc'))
    # Pass site url to spider doing start requests
    spider_name = "testspider_startrequests.py"
    spider_filename = os.path.join(directory, "testproject", "spider_templates", spider_name)
    spider_target_place = os.path.join(directory, "testproject", "spiders", spider_name)
    if not site:
        return
    with open(spider_filename) as spider_file:
        spider_string = spider_file.read().format(site.url("page1.html"), site.url("page2.html"))
        with open(spider_target_place, "wb") as file_target:
            file_target.write(spider_string.encode('utf8'))



# -*- coding: utf-8 -*-
from scrapy.settings import Settings

from . import settings


def get_scrapyrt_settings(log_file=None):
    spider_settings = {
        "LOG_LEVEL": "DEBUG",
        "LOG_ENABLED": bool(log_file),
        "LOG_FILE": log_file,
        "LOG_STDOUT": False,
        "EXTENSIONS": {
            'scrapy.extensions.logstats.LogStats': None,
            'scrapy.webservice.WebService': None,
            'scrapy.extensions.telnet.TelnetConsole': None,
            'scrapy.extensions.throttle.AutoThrottle': None
        }
    }
    return spider_settings


def get_project_settings(module=None, custom_settings=None):
    crawler_settings = Settings()
    if module is None:
        module = settings.PROJECT_SETTINGS
    crawler_settings.setmodule(module, priority='project')
    if custom_settings:
        assert isinstance(custom_settings, dict)
        crawler_settings.setdict(custom_settings, priority='cmdline')
    return crawler_settings



# -*- coding: utf-8 -*-
from collections import OrderedDict
from copy import deepcopy
import datetime
import os
import six

from scrapy import signals
from scrapy.crawler import CrawlerRunner, Crawler
from scrapy.exceptions import DontCloseSpider
from scrapy.http import Request
from twisted.web.error import Error
from twisted.internet import defer

from . import log
from .conf import settings
from .conf.spider_settings import get_scrapyrt_settings, get_project_settings
from .decorators import deprecated
from .log import setup_spider_logging


class ScrapyrtCrawler(Crawler):
    """Main and only difference from base class -
    ScrapyrtCrawler allows us to call or not call
    start_requests.

    https://github.com/scrapy/scrapy/blob/master/scrapy/crawler.py#L52

    TODO: PR to scrapy - ability to set start_requests here.

    """
    def __init__(self, spidercls, crawler_settings, start_requests=False):
        super(ScrapyrtCrawler, self).__init__(spidercls, crawler_settings)
        self.start_requests = start_requests

    @defer.inlineCallbacks
    def crawl(self, *args, **kwargs):
        assert not self.crawling, "Crawling already taking place"
        self.crawling = True
        try:
            self.spider = self._create_spider(*args, **kwargs)
            self.engine = self._create_engine()
            if self.start_requests:
                start_requests = iter(self.spider.start_requests())
            else:
                start_requests = ()
            yield self.engine.open_spider(self.spider, start_requests)
            yield defer.maybeDeferred(self.engine.start)
        except Exception:
            self.crawling = False
            raise


class ScrapyrtCrawlerProcess(CrawlerRunner):

    def __init__(self, settings, scrapyrt_manager):
        super(ScrapyrtCrawlerProcess, self).__init__(settings)
        self.scrapyrt_manager = scrapyrt_manager

    def crawl(self, spidercls, *args, **kwargs):
        if isinstance(spidercls, six.string_types):
            spidercls = self.spider_loader.load(spidercls)
        # creating our own crawler that will allow us to disable start requests easily
        crawler = ScrapyrtCrawler(
            spidercls, self.settings, self.scrapyrt_manager.start_requests)
        self.scrapyrt_manager.crawler = crawler
        # Connecting signals to handlers that control crawl process
        crawler.signals.connect(self.scrapyrt_manager.get_item,
                                signals.item_scraped)
        crawler.signals.connect(self.scrapyrt_manager.collect_dropped,
                                signals.item_dropped)
        crawler.signals.connect(self.scrapyrt_manager.spider_idle,
                                signals.spider_idle)
        crawler.signals.connect(self.scrapyrt_manager.handle_spider_error,
                                signals.spider_error)
        crawler.signals.connect(self.scrapyrt_manager.handle_scheduling,
                                signals.request_scheduled)
        dfd = super(ScrapyrtCrawlerProcess, self).crawl(crawler, *args, **kwargs)
        _cleanup_handler = setup_spider_logging(crawler.spider, self.settings)

        def cleanup_logging(result):
            _cleanup_handler()
            return result

        return dfd.addBoth(cleanup_logging)


class CrawlManager(object):
    """
    Runs crawls
    """

    def __init__(self, spider_name, request_kwargs, max_requests=None, start_requests=False):
        self.spider_name = spider_name
        self.log_dir = settings.LOG_DIR
        self.items = []
        self.items_dropped = []
        self.errors = []
        self.max_requests = int(max_requests) if max_requests else None
        self.timeout_limit = int(settings.TIMEOUT_LIMIT)
        self.request_count = 0
        self.debug = settings.DEBUG
        self.crawler_process = None
        self.crawler = None
        # callback will be added after instantiation of crawler object
        # because we need to know if spider has method available
        self.callback_name = request_kwargs.pop('callback', None) or 'parse'
        if request_kwargs.get("url"):
            self.request = self.create_spider_request(deepcopy(request_kwargs))
        else:
            self.request = None
        self.start_requests = start_requests
        self._request_scheduled = False

    def crawl(self, *args, **kwargs):
        self.crawler_process = ScrapyrtCrawlerProcess(
            self.get_project_settings(), self)
        try:
            dfd = self.crawler_process.crawl(self.spider_name, *args, **kwargs)
        except KeyError as e:
            # Spider not found.
            raise Error('404', message=str(e))
        dfd.addCallback(self.return_items)
        return dfd

    def _get_log_file_path(self):
        log_dir = os.path.join(self.log_dir, self.spider_name)
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        time_format = settings.SPIDER_LOG_FILE_TIMEFORMAT
        filename = datetime.datetime.now().strftime(time_format) + '.log'
        return os.path.join(log_dir, filename)

    def get_project_settings(self):
        # set logfile for a job
        log_file = self._get_log_file_path()
        custom_settings = get_scrapyrt_settings(log_file=log_file)
        return get_project_settings(custom_settings=custom_settings)

    @deprecated(use_instead='.crawl()')
    def create_crawler(self, **kwargs):
        return self.crawl()

    def spider_idle(self, spider):
        """Handler of spider_idle signal.

        Schedule request for url given to api, with optional callback
        that can be passed as GET parameter.

        spider_idle signal is used because we want to optionally enable
        start_requests for the spider and if request is scheduled in
        spider_opened signal handler it's fired earlier then start_requests
        which is totally wrong.

        """
        if spider is self.crawler.spider and self.request and not self._request_scheduled:
            callback = getattr(self.crawler.spider, self.callback_name)
            assert callable(callback), 'Invalid callback'
            self.request = self.request.replace(callback=callback)
            modify_request = getattr(
                self.crawler.spider, "modify_realtime_request", None)
            if callable(modify_request):
                self.request = modify_request(self.request)
            spider.crawler.engine.crawl(self.request, spider)
            self._request_scheduled = True
            raise DontCloseSpider

    def handle_scheduling(self, request, spider):
        """Handler of request_scheduled signal.

        For every scheduled request check if number of requests is less
        then limit and runtime doesn't exceed limit as well.

        """
        if spider is self.crawler.spider:
            self.limit_requests(spider)
            self.limit_runtime(spider)

    def limit_runtime(self, spider):
        """Stop crawl if it takes too long."""
        start_time = self.crawler.stats.get_value("start_time")
        time_now = datetime.datetime.utcnow()
        if (time_now - start_time).seconds >= self.timeout_limit:
            spider.crawler.engine.close_spider(spider, reason="timeout")

    def limit_requests(self, spider):
        """Stop crawl after reaching max_requests."""
        if self.max_requests and self.max_requests <= self.request_count:
            reason = "stop generating requests, only {} requests allowed".format(
                self.max_requests)
            spider.crawler.engine.close_spider(spider, reason=reason)
        else:
            self.request_count += 1

    def handle_spider_error(self, failure, spider):
        if spider is self.crawler.spider and self.debug:
            fail_data = failure.getTraceback()
            self.errors.append(fail_data)

    def get_item(self, item, response, spider):
        if spider is self.crawler.spider:
            self.items.append(item)

    def collect_dropped(self, item, response, exception, spider):
        if spider is self.crawler.spider:
            self.items_dropped.append({
                "item": item,
                "exception": str(exception),
                "response": response
            })

    def return_items(self, result):
        stats = self.crawler.stats.get_stats()
        stats = OrderedDict((k, v) for k, v in sorted(stats.items()))
        results = {
            "items": self.items,
            "items_dropped": self.items_dropped,
            "stats": stats,
            "spider_name": self.spider_name,
        }
        if self.debug:
            results["errors"] = self.errors
        return results

    def create_spider_request(self, kwargs):
        url = kwargs.pop('url')
        try:
            req = Request(url, **kwargs)
        except (TypeError, ValueError) as e:
            msg = "Error while creating Scrapy Request, {}"
            message = msg.format(str(e))
            raise Error('400', message=message)

        req.dont_filter = True
        msg = u"Created request for spider {} with url {} and kwargs {}"
        msg = msg.format(self.spider_name, url, repr(kwargs))
        log.msg(msg)
        return req



# -*- coding: utf-8 -*-
import os

TESTS_PATH = os.path.realpath(os.path.dirname(__file__))
PROJECT_PATH = os.path.realpath(os.path.join(TESTS_PATH, '..'))
SAMPLE_DATA = os.path.join(TESTS_PATH, 'sample_data')



# -*- coding: utf-8 -*-
from mock import patch
from twisted.trial import unittest

from scrapyrt.conf import Settings

from . import default_settings


class TestSettings(unittest.TestCase):

    @patch('scrapyrt.conf.default_settings', default_settings)
    def setUp(self):
        self.settings = Settings()

    def test_getattr(self):
        self.assertEqual(self.settings.A, 'A')
        self.assertEqual(self.settings.TEST, [1, 2, 3])

        # invalid (or hidden in this way) settings should not be visible
        self.assertRaises(AttributeError, getattr, self.settings, '_HIDDEN')
        self.assertRaises(AttributeError, getattr, self.settings, 'hidden')
        self.assertRaises(AttributeError, getattr, self.settings, 'HiDdEn')

    def test_setmodule(self):
        from . import settings
        self.assertEqual(self.settings.A, 'A')
        self.settings.setmodule(settings)
        self.assertEqual(self.settings.A, 'B')
        self.assertEqual(self.settings.TEST, [1, 2, 3])

    def test_setmodule_string(self):
        self.assertEqual(self.settings.A, 'A')
        self.settings.setmodule('tests.test_settings.settings')
        self.assertEqual(self.settings.A, 'B')
        self.assertEqual(self.settings.TEST, [1, 2, 3])

    def test_set(self):
        self.assertEqual(self.settings.A, 'A')
        self.settings.set('A', 'C')
        self.assertEqual(self.settings.A, 'C')

        self.assertEqual(self.settings.TEST, [1, 2, 3])
        self.settings.set('TEST', [])
        self.assertEqual(self.settings.TEST, [])

        # invalid setting names
        self.settings.set('_HIDDEN', 1)
        self.assertRaises(AttributeError, getattr, self.settings, '_HIDDEN')
        self.settings.set('hidden', 1)
        self.assertRaises(AttributeError, getattr, self.settings, 'hidden')
        self.settings.set('HiDdEn', 1)
        self.assertRaises(AttributeError, getattr, self.settings, 'HiDdEn')

    def test_freeze(self):
        self.assertEqual(self.settings.A, 'A')
        self.settings.set('A', 'D')
        self.assertEqual(self.settings.A, 'D')
        self.assertFalse(self.settings.frozen)
        self.settings.freeze()
        self.assertTrue(self.settings.frozen)
        self.assertRaises(TypeError, self.settings.set, 'A', 'E')



import re

import pytest

from scrapyrt.utils import extract_scrapy_request_args


class TestUtils(object):
    def test_get_scrapy_request_args(self):
        args = {
            "url": "http://foo.com",
            "callback": "parse",
            "noise": True
        }

        result = extract_scrapy_request_args(args)

        assert result["url"] == "http://foo.com"
        assert result["callback"] == "parse"
        assert "noise" not in result

    def test_get_scrapy_request_args_error(self):
        args = {
            "url": "http://foo.com",
            "callback": "parse",
            "noise": True
        }

        with pytest.raises(ValueError) as e:
            extract_scrapy_request_args(args, raise_error=True)

        expected_msg =u"'noise' is not a valid argument for scrapy.Request"
        assert re.search(expected_msg, str(e.value))




# -*- coding: utf-8 -*-
import logging
import os
import sys
from logging.config import dictConfig

from scrapy.settings import Settings
from scrapy.utils.log import DEFAULT_LOGGING, TopLevelFormatter
from twisted.python import log
from twisted.python.log import startLoggingWithObserver
from twisted.python.logfile import DailyLogFile

from .conf import settings as scrapyrt_settings
from .utils import to_bytes

DEBUG = logging.DEBUG
INFO = logging.INFO
WARNING = logging.WARNING
ERROR = logging.ERROR
CRITICAL = logging.CRITICAL
SILENT = CRITICAL + 1


def msg(message=None, **kwargs):
    kwargs['logLevel'] = kwargs.pop('level', INFO)
    kwargs.setdefault('system', 'scrapyrt')
    if message is None:
        log.msg(**kwargs)
    else:
        log.msg(message, **kwargs)


def err(_stuff=None, _why=None, **kwargs):
    kwargs['logLevel'] = kwargs.pop('level', ERROR)
    kwargs.setdefault('system', 'scrapyrt')
    log.err(_stuff, _why, **kwargs)


class ScrapyrtFileLogObserver(log.FileLogObserver):

    def __init__(self, f, encoding='utf-8'):
        self.encoding = encoding.lower()
        log.FileLogObserver.__init__(self, f)

    def _adapt_eventdict(self, event_dict):
        """Adapt event dict making it suitable for logging with Scrapyrt log
        observer.

        :return: adapted event_dict, None if message should be ignored.

        """
        if event_dict.get('category') == 'scrapy.exceptions.ScrapyDeprecationWarning':
            # ignore Scrapy deprecation warning in ScrapyRT log
            return
        if event_dict.get('system') == 'scrapy':
            return
        if ('HTTPChannel' in event_dict.get('system') and
                'Log opened.' in event_dict.get('message', '')):
            # useless log message caused by scrapy.log.start
            return
        return event_dict

    def _unicode_to_str(self, eventDict):
        message = eventDict.get('message')
        if message:
            eventDict['message'] = tuple(
                to_bytes(x, self.encoding) for x in message)
        return eventDict

    def emit(self, eventDict):
        eventDict = self._adapt_eventdict(eventDict)
        if eventDict is None:
            return
        eventDict = self._unicode_to_str(eventDict)
        log.FileLogObserver.emit(self, eventDict)


class SpiderFilter(logging.Filter):
    """Filter messages from other spiders and undefined loggers.

    Accept messages that have 'spider' key in extra and it matches given spider.

    """

    def __init__(self, spider):
        self.spider = spider

    def filter(self, record):
        spider = getattr(record, 'spider', None)
        return spider and spider is self.spider


def setup_logging():
    if not os.path.exists(scrapyrt_settings.LOG_DIR):
        os.makedirs(scrapyrt_settings.LOG_DIR)
    if scrapyrt_settings.LOG_FILE:
        logfile = DailyLogFile.fromFullPath(
            os.path.join(scrapyrt_settings.LOG_DIR,
                         scrapyrt_settings.LOG_FILE)
        )
    else:
        logfile = sys.stderr
    observer = ScrapyrtFileLogObserver(logfile, scrapyrt_settings.LOG_ENCODING)
    startLoggingWithObserver(observer.emit, setStdout=False)

    # setup general logging for Scrapy
    if not sys.warnoptions:
        # Route warnings through python logging
        logging.captureWarnings(True)

    observer = log.PythonLoggingObserver('twisted')
    observer.start()
    logging.root.setLevel(logging.NOTSET)
    dictConfig(DEFAULT_LOGGING)


def setup_spider_logging(spider, settings):
    """Initialize and configure default loggers

    Copied from Scrapy and updated, because version from Scrapy:

     1) doesn't close handlers and observers
     2) opens logobserver for twisted logging each time it's called -
        you can find N log observers logging the same message N
        after N crawls.

    so there's no way to reuse it.

    :return: method that should be called to cleanup handler.

    """
    if isinstance(settings, dict):
        settings = Settings(settings)

    # Looging stdout is a bad idea when mutiple crawls are running
    # if settings.getbool('LOG_STDOUT'):
    #     sys.stdout = StreamLogger(logging.getLogger('stdout'))
    filename = settings.get('LOG_FILE')
    if filename:
        encoding = settings.get('LOG_ENCODING')
        handler = logging.FileHandler(filename, encoding=encoding)
    elif settings.getbool('LOG_ENABLED'):
        handler = logging.StreamHandler()
    else:
        handler = logging.NullHandler()
    formatter = logging.Formatter(
        fmt=settings.get('LOG_FORMAT'),
        datefmt=settings.get('LOG_DATEFORMAT')
    )
    handler.setFormatter(formatter)
    handler.setLevel(settings.get('LOG_LEVEL'))
    filters = [
        TopLevelFormatter(['scrapy']),
        SpiderFilter(spider),
    ]
    for _filter in filters:
        handler.addFilter(_filter)
    logging.root.addHandler(handler)

    _cleanup_functions = [
        lambda: [handler.removeFilter(f) for f in filters],
        lambda: logging.root.removeHandler(handler),
        handler.close,
    ]

    def cleanup():
        for func in _cleanup_functions:
            try:
                func()
            except Exception as e:
                err(e)

    return cleanup



# -*- coding: utf-8 -*-
import os
from time import sleep
import datetime

import pytest
from mock import patch, MagicMock
from scrapy import Item
from scrapy.exceptions import DontCloseSpider
from scrapy.http import Response
from scrapy.settings import Settings
from scrapy.utils.test import get_crawler
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure
from twisted.trial import unittest
from twisted.web.error import Error

from scrapyrt.core import CrawlManager
from scrapyrt.conf import settings

from .spiders import MetaSpider


class TestCrawlManager(unittest.TestCase):

    def setUp(self):
        self.url = 'http://localhost'
        self.kwargs = {'url': self.url, 'dont_filter': True}
        self.crawler = MagicMock()
        self.spider = MetaSpider.from_crawler(self.crawler)
        self.crawler.spider = self.spider
        self.crawl_manager = self._create_crawl_manager()
        self.crawl_manager.crawler = self.crawler
        self.item = Item()
        self.response = Response('http://localhost')
        self.another_spider = MetaSpider.from_crawler(self.crawler)

    def _create_crawl_manager(self):
        crawl_manager = CrawlManager(self.spider.name, self.kwargs.copy())
        crawl_manager.crawler = self.crawler
        return crawl_manager


@patch('scrapyrt.core.ScrapyrtCrawlerProcess.crawl', return_value=Deferred())
class TestCrawl(TestCrawlManager):

    def test_crawl(self, crawler_process_mock):
        result = self.crawl_manager.crawl()
        self.assertIsInstance(result, Deferred)
        self.assertGreater(len(result.callbacks), 0)
        self.assertEqual(
            result.callbacks[0][0][0], self.crawl_manager.return_items)

    def test_no_spider(self, crawler_process_mock):
        # spider wasn't found
        crawler_process_mock.side_effect = KeyError
        exception = self.assertRaises(
            Error, self.crawl_manager.crawl)
        self.assertTrue(crawler_process_mock.called)
        self.assertEqual(exception.status, '404')

    def test_spider_exists(self, crawler_process_mock):
        result = self.crawl_manager.crawl()
        self.assertTrue(crawler_process_mock.called)
        self.assertIs(result, crawler_process_mock.return_value)

    def test_spider_arguments_are_passed(self, crawler_process_mock):
        spider_args = ['a', 'b']
        spider_kwargs = {'a': 1, 'b': 2}
        self.crawl_manager.crawl(*spider_args, **spider_kwargs)
        self.assertTrue(crawler_process_mock.called)
        call_args, call_kwargs = crawler_process_mock.call_args
        for arg in spider_args:
            self.assertIn(arg, call_args)
        self.assertDictContainsSubset(spider_kwargs, call_kwargs)


class TestGetProjectSettings(TestCrawlManager):

    def test_get_project_settings(self):
        result = self.crawl_manager.get_project_settings()
        self.assertIsInstance(result, Settings)


class TestSpiderIdle(TestCrawlManager):

    def setUp(self):
        super(TestSpiderIdle, self).setUp()
        self.crawler.spider = self.spider
        # test callback
        self.spider.parse_something = lambda: None
        self.crawl_manager.callback_name = 'parse_something'
        self.request = self.crawl_manager.request

    def _call_spider_idle(self):
        try:
            self.crawl_manager.spider_idle(self.spider)
        except DontCloseSpider:
            pass

    def test_spider_opened(self):
        self.assertIsNone(self.crawl_manager.request.callback)
        self._call_spider_idle()
        self.crawler.engine.crawl.assert_called_once_with(
            self.crawl_manager.request, self.spider)
        self.assertNotEqual(self.request, self.crawl_manager.request)
        self.assertEquals(
            self.crawl_manager.request.callback, self.spider.parse_something)

    def test_raise_error_if_not_callable(self):
        self.spider.parse_something = None
        self.assertRaises(
            AssertionError, self.crawl_manager.spider_idle, self.spider)
        self.assertFalse(self.crawler.engine.crawl.called)

    def test_modify_realtime_request(self):
        self.assertDictEqual(self.crawl_manager.request.meta, {})
        self.assertEqual(self.crawl_manager.request.method, 'GET')

        def modify_realtime_request(request):
            request = request.replace(method='POST')
            request.meta['foo'] = 'bar'
            return request

        self.spider.modify_realtime_request = modify_realtime_request
        self._call_spider_idle()
        self.crawler.engine.crawl.assert_called_once_with(
            self.crawl_manager.request, self.spider)
        self.assertEqual(self.crawl_manager.request.method, 'POST')
        self.assertEqual(self.crawl_manager.request.meta['foo'], 'bar')

    def test_modify_realtime_request_is_not_callable(self):
        self.spider.modify_realtime_request = None
        self._call_spider_idle()
        self.crawler.engine.crawl.assert_called_once_with(
            self.crawl_manager.request, self.spider)
        self.assertNotEqual(self.request, self.crawl_manager.request)


class TestHandleScheduling(TestCrawlManager):

    def setUp(self):
        super(TestHandleScheduling, self).setUp()
        self.crawl_manager.limit_requests = MagicMock()
        self.crawl_manager.limit_runtime = MagicMock()

    def test_handle_scheduling(self):
        self.crawl_manager.handle_scheduling(
            self.crawl_manager.request, self.spider)
        self.crawl_manager.limit_requests.assert_called_once_with(self.spider)
        self.crawl_manager.limit_runtime.assert_called_once_with(self.spider)

    def test_handle_scheduling_another_spider(self):
        self.crawl_manager.handle_scheduling(
            self.crawl_manager.request, self.another_spider)
        self.assertFalse(self.crawl_manager.limit_requests.called)
        self.assertFalse(self.crawl_manager.limit_runtime.called)


class TestLimitRuntime(TestCrawlManager):

    def setUp(self):
        super(TestLimitRuntime, self).setUp()
        self.crawl_manager.timeout_limit = 1
        self.crawler.stats.get_value.return_value = datetime.datetime.utcnow()

    def _test_limit_runtime(self):
        self.crawl_manager.limit_runtime(self.spider)
        self.assertFalse(self.crawler.engine.close_spider.called)
        sleep(1)
        self.crawl_manager.limit_runtime(self.spider)
        self.assertTrue(self.crawler.engine.close_spider.called)

    def test_limit_runtime(self):
        self._test_limit_runtime()

    def test_string_number_timeout_value(self):
        _timeout = settings.TIMEOUT_LIMIT
        try:
            settings.TIMEOUT_LIMIT = '1'
            self.crawl_manager = self._create_crawl_manager()
            self._test_limit_runtime()
        finally:
            settings.TIMEOUT_LIMIT = _timeout

    def test_wrong_timeout_value(self):
        _timeout = settings.TIMEOUT_LIMIT
        try:
            settings.TIMEOUT_LIMIT = 'foo'
            self.assertRaises(
                ValueError, CrawlManager, self.spider.name, self.kwargs.copy())
        finally:
            settings.TIMEOUT_LIMIT = _timeout


class TestHandleSpiderError(TestCrawlManager):

    def setUp(self):
        super(TestHandleSpiderError, self).setUp()
        self.exception_message = 'Foo'
        self.exception = Exception(self.exception_message)
        self.failure = Failure(self.exception)

    def test_handle_spider_error_debug_true(self):
        self.assertEqual(len(self.crawl_manager.errors), 0)
        self.crawl_manager.handle_spider_error(self.failure, self.spider)
        self.assertEqual(len(self.crawl_manager.errors), 1)
        self.assertIn('Traceback', self.crawl_manager.errors[0])
        self.assertIn(self.exception.__class__.__name__,
                      self.crawl_manager.errors[0])
        self.assertIn(self.exception_message, self.crawl_manager.errors[0])

    def test_handle_spider_error_debug_false(self):
        self.crawl_manager.debug = False
        self.assertEqual(len(self.crawl_manager.errors), 0)
        self.crawl_manager.handle_spider_error(self.failure, self.spider)
        self.assertEqual(len(self.crawl_manager.errors), 0)

    def test_handle_spider_error_another_spider(self):
        self.assertEqual(len(self.crawl_manager.errors), 0)
        self.crawl_manager.handle_spider_error(
            self.failure, self.another_spider)
        self.assertEqual(len(self.crawl_manager.errors), 0)


class TestLimitRequests(TestCrawlManager):

    def test_max_requests_not_set(self):
        for i in range(100):
            self.crawl_manager.limit_requests(self.spider)
        self.assertFalse(self.crawler.engine.close_spider.called)

    def test_max_requests_set(self):
        self.crawl_manager.max_requests = 10
        for i in range(self.crawl_manager.max_requests):
            self.crawl_manager.limit_requests(self.spider)
        self.assertFalse(self.crawler.engine.close_spider.called)
        self.crawl_manager.limit_requests(self.spider)
        self.assertTrue(self.crawler.engine.close_spider.called)


class TestGetItem(TestCrawlManager):

    def setUp(self):
        super(TestGetItem, self).setUp()
        self.item = Item()

    def test_get_item(self):
        self.assertEqual(len(self.crawl_manager.items), 0)
        self.crawl_manager.get_item(self.item, self.response, self.spider)
        self.assertEqual(len(self.crawl_manager.items), 1)
        self.assertEqual(self.crawl_manager.items[0], self.item)

    def test_get_item_another_spider(self):
        self.assertEqual(len(self.crawl_manager.items), 0)
        self.crawl_manager.get_item(
            self.item, self.response, self.another_spider)
        self.assertEqual(len(self.crawl_manager.items), 0)


class TestCollectDropped(TestCrawlManager):

    def setUp(self):
        super(TestCollectDropped, self).setUp()
        self.exception = Exception('foo')
        self.expected_result = {
            'item': self.item,
            'response': self.response,
            'exception': str(self.exception)
        }

    def test_collect_dropped(self):
        self.assertEqual(len(self.crawl_manager.items_dropped), 0)
        self.crawl_manager.collect_dropped(
            self.item, self.response, self.exception, self.spider)
        self.assertEqual(len(self.crawl_manager.items_dropped), 1)
        self.assertEqual(len(self.crawl_manager.items_dropped), 1)
        self.assertEqual(
            self.crawl_manager.items_dropped[0], self.expected_result)

    def test_collect_dropped_another_spider(self):
        self.assertEqual(len(self.crawl_manager.items_dropped), 0)
        self.crawl_manager.collect_dropped(
            self.item, self.response, self.exception, self.another_spider)
        self.assertEqual(len(self.crawl_manager.items_dropped), 0)


class TestReturnItems(TestCrawlManager):

    def setUp(self):
        super(TestReturnItems, self).setUp()
        self.stats = {
            'log_count/INFO': 6,
            'scheduler/enqueued/memory': 4,
            'scheduler/dequeued/memory': 4,
        }
        self.crawl_manager.crawler = MagicMock()
        self.crawl_manager.crawler.stats.get_stats.return_value = self.stats
        self.expected_result = {
            'items': self.crawl_manager.items,
            'items_dropped': self.crawl_manager.items_dropped,
            'stats': self.stats.copy(),
            'spider_name': self.spider.name,
        }

    def test_return_items(self):
        result = self.crawl_manager.return_items(None)
        self.assertDictContainsSubset(self.expected_result, result)
        self.assertEqual(list(sorted(self.stats.keys())), list(result['stats'].keys()))
        # debug = True by default
        self.assertIn('errors', result)
        self.assertEquals(result['errors'], self.crawl_manager.errors)

    def test_return_items_without_debug(self):
        self.crawl_manager.debug = False
        result = self.crawl_manager.return_items(None)
        self.assertDictEqual(self.expected_result, result)
        self.assertNotIn('errors', result)


class TestCreateSpiderRequest(TestCrawlManager):

    def test_valid_arguments(self):
        req = self.crawl_manager.create_spider_request(self.kwargs)
        self.assertTrue(req.dont_filter)
        self.assertEqual(req.url, self.url)

    def test_invalid_arguments(self):
        self.kwargs['url1'] = 'http://localhost/foo'
        exception = self.assertRaises(
            Error, self.crawl_manager.create_spider_request, self.kwargs)
        self.assertEqual(exception.status, '400')

    def test_invalid_url(self):
        self.kwargs['url'] = '//localhost/foo'
        exception = self.assertRaises(
            Error, self.crawl_manager.create_spider_request, self.kwargs)
        self.assertEqual(exception.status, '400')
        self.kwargs['url'] = 'localhost/foo'
        exception = self.assertRaises(
            Error, self.crawl_manager.create_spider_request, self.kwargs)
        self.assertEqual(exception.status, '400')


class TestStartRequests(unittest.TestCase):

    def setUp(self):
        self.url = 'http://localhost'
        self.kwargs = {'url': self.url}
        self.start_requests_mock = MagicMock()
        self.spidercls = MetaSpider
        self._start_requests = self.spidercls.start_requests
        self.spidercls.start_requests = self.start_requests_mock
        self.crawler = get_crawler(self.spidercls)

        class CustomCrawlManager(CrawlManager):

            def get_project_settings(self):
                crawl_settings = super(
                    CustomCrawlManager, self).get_project_settings()
                crawl_settings.setdict(
                    {'SPIDER_MODULES': 'tests.spiders'}, priority='cmdline')
                return crawl_settings

        self.crawl_manager = CustomCrawlManager(
            self.spidercls.name, self.kwargs.copy())
        self.crawl_manager.crawler = self.crawler

    def tearDown(self):
        self.spidercls.start_requests = self._start_requests

    @patch('scrapy.crawler.ExecutionEngine')
    def test_start_requests_true(self, _):
        self.crawl_manager.start_requests = True
        self.crawl_manager.crawl()
        self.assertEqual(self.start_requests_mock.call_count, 1)

    @patch('scrapy.crawler.ExecutionEngine')
    def test_start_requests_false(self, _):
        self.crawl_manager.start_requests = False
        self.crawl_manager.crawl()
        self.assertEqual(self.start_requests_mock.call_count, 0)


class TestCreateProperLogFile(TestCrawlManager):
    def test_filename(self):
        logdir = "some_dir_name"
        self.crawl_manager.log_dir = logdir
        path = self.crawl_manager._get_log_file_path()
        filename = os.path.basename(path)
        expected_format = '%Y-%m-%dT%H%M%S.%f.log'
        datetime_object = datetime.datetime.strptime(filename, expected_format)
        now = datetime.datetime.now()
        assert datetime_object
        delta = now - datetime_object
        assert delta.seconds < 60



import inspect

import sys

import six
from scrapy import Request


def extract_scrapy_request_args(dictionary, raise_error=False):
    """
    :param dictionary: Dictionary with parameters passed to API
    :param raise_error: raise ValueError if key is not valid arg for
                        scrapy.http.Request
    :return: dictionary of valid scrapy.http.Request positional and keyword
            arguments.
    """
    result = dictionary.copy()
    args = inspect.getargspec(Request.__init__).args
    for key in dictionary.keys():
        if key not in args:
            result.pop(key)
            if raise_error:
                msg = u"{!r} is not a valid argument for scrapy.Request.__init__"
                raise ValueError(msg.format(key))
    return result


try:
    from scrapy.utils.python import to_bytes
except ImportError:
    def to_bytes(text, encoding=None, errors='strict'):
        """Return the binary representation of `text`. If `text`
        is already a bytes object, return it as-is."""
        if isinstance(text, bytes):
            return text
        if not isinstance(text, six.string_types):
            raise TypeError('to_bytes must receive a unicode, str or bytes '
                            'object, got %s' % type(text).__name__)
        if encoding is None:
            encoding = 'utf-8'
        return text.encode(encoding, errors)



# -*- coding: utf-8 -*-
from mock import MagicMock
from scrapy import signals
from twisted.internet.defer import Deferred
from twisted.trial import unittest

from scrapyrt.core import CrawlManager, ScrapyrtCrawlerProcess

from .spiders import MetaSpider
from .utils import get_settings


class CralwerProcessTestCase(unittest.TestCase):

    def _mock_method(self, obj, method):
        msg = "can't mock, class {} doesn't have method {}".format(
            obj.__class__.__name__, method)
        assert hasattr(obj, method), msg
        setattr(obj, method, MagicMock(spec=lambda: None))

    def test_signals(self):
        """Need to be sure that all signals are bind to appropriate handlers
        right after crawler is created.

        """
        crawl_manager = CrawlManager('test', {'url': 'http://localhost'})

        signals_and_handlers = [
            ('item_scraped', 'get_item'),
            ('item_dropped', 'collect_dropped'),
            ('spider_idle', 'spider_idle'),
            ('spider_error', 'handle_spider_error'),
            ('request_scheduled', 'handle_scheduling'),
        ]
        for _, handler in signals_and_handlers:
            self._mock_method(crawl_manager, handler)
        settings = get_settings()
        crawler_process = ScrapyrtCrawlerProcess(settings, crawl_manager)
        dfd = crawler_process.crawl(MetaSpider)
        self.assertIsInstance(dfd, Deferred)
        crawler = crawl_manager.crawler
        for signal, handler in signals_and_handlers:
            crawler.signals.send_catch_log(
                signal=getattr(signals, signal), spider=crawler.spider)
            handler_mock = getattr(crawl_manager, handler)
            self.assertEquals(handler_mock.call_count, 1)



# -*- coding: utf-8 -*-
import json

import pytest
import re
from mock import MagicMock, patch, Mock
from twisted.trial import unittest
from twisted.web.error import Error
import requests
from twisted.web.server import Request

from scrapyrt.resources import CrawlResource

from .servers import ScrapyrtTestServer, MockServer


@pytest.fixture()
def server(request):
    target_site = MockServer()
    target_site.start()
    server = ScrapyrtTestServer(site=target_site)

    def close():
        server.stop()
        target_site.stop()

    request.addfinalizer(close)
    server.target_site = target_site
    server.start()
    return server


@pytest.fixture()
def t_req():
    return MagicMock(spec=Request)


@pytest.fixture()
def resource():
    return CrawlResource()


class TestCrawlResource(object):

    def test_is_leaf(self):
        assert CrawlResource.isLeaf

    def test_render_GET(self, t_req, resource):
        t_req.args = {
            b'url': [b'http://foo'],
            b'spider_name': [b'test']
        }
        resource.validate_options = Mock()

        with patch('scrapyrt.core.CrawlManager', spec=True) as manager:
            instance = manager.return_value
            resource.render_GET(t_req)
        scrapy_params = {'url': 'http://foo'}
        api_params = {'spider_name': 'test', 'url': 'http://foo'}
        resource.validate_options.assert_called_once_with(
            scrapy_params, api_params
        )
        assert instance.crawl.called

    def test_render_POST(self, t_req, resource):
        t_req.content.getvalue.return_value = json.dumps({
                'spider_name': 'test',
                'request': {
                    'url': 'http://foo.com'
                }
        })
        resource.validate_options = Mock()
        with patch('scrapyrt.core.CrawlManager', spec=True) as manager:
            instance = manager.return_value
            resource.render_POST(t_req)

        scrapy_params = {'url': 'http://foo.com'}
        api_params = {
            'request': {'url': 'http://foo.com'},
            'spider_name': 'test'
        }
        assert instance.crawl.called
        resource.validate_options.assert_called_once_with(
            scrapy_params, api_params
        )

    def test_render_POST_invalid_json(self, t_req, resource):
        t_req.content.getvalue.return_value = b'{{{{{'
        with patch('scrapyrt.core.CrawlManager', spec=True) as manager:
            with pytest.raises(Error) as e:
                resource.render_POST(t_req)
        assert e.value.status == '400'
        assert re.search('Invalid JSON in POST', e.value.message)
        assert not manager.return_value.crawl.called

    def test_render_POST_invalid_options(self, t_req, resource):
        t_req.content.getvalue.return_value = json.dumps({
            'spider_name': 'tests',
            'request': {
                'foo': 'bar'
            }
        })
        resource.validate_options = Mock()
        with patch('scrapyrt.core.CrawlManager', spec=True):
            with pytest.raises(Error) as e:
                resource.render_POST(t_req)
        assert e.value.status == '400'
        msg = "'foo' is not a valid argument"
        assert re.search(msg, e.value.message)

    @pytest.mark.parametrize('scrapy_args,api_args,has_error', [
        ({'url': 'aa'}, {}, False),
        ({}, {}, True)
    ])
    def test_validate_options(self, resource,
                              scrapy_args, api_args, has_error):
        if has_error:
            with pytest.raises(Error) as e:
                resource.validate_options(scrapy_args, api_args)
            assert e.value.status == '400'
            assert re.search("\'url\' is required", e.value.message)
        else:
            result = resource.validate_options(scrapy_args, api_args)
            assert result is None

    def test_prepare_response(self, resource):
        result = {
            'items': [1, 2],
            'stats': [99],
            'spider_name': 'test'
        }
        prepared_res = resource.prepare_response(result)
        expected = [
            ('status', 'ok'),
            ('items', [1, 2]),
            ('items_dropped', []),
            ('stats', [99]),
            ('spider_name', 'test')
        ]
        for key, value in expected:
            assert prepared_res[key] == value


class TestCrawlResourceGetRequiredArgument(unittest.TestCase):

    def setUp(self):
        self.resource = CrawlResource()
        self.url = 'http://localhost:1234'
        self.data = {'url': self.url}

    def test_get_argument(self):
        self.assertEqual(
            self.resource.get_required_argument(self.data, 'url'), self.url)

    def test_raise_error(self):
        exception = self.assertRaises(
            Error, self.resource.get_required_argument, self.data, 'key')
        self.assertEqual(exception.status, '400')

    def test_empty_argument(self):
        self.data['url'] = ''
        exception = self.assertRaises(
            Error, self.resource.get_required_argument, self.data, 'url')
        self.assertEqual(exception.status, '400')


def perform_get(url, api_params, spider_data):
    api_params.update(spider_data)
    return requests.get(url, params=api_params)


def perform_post(url, api_params, spider_data):
    post_data = {"request": spider_data}
    post_data.update(api_params)
    return requests.post(url, json=post_data)


class TestCrawlResourceIntegration(object):
    @pytest.mark.parametrize("method", [
        perform_get, perform_post
    ])
    def test_no_parameters(self, server, method):
        res = method(server.url('crawl.json'), {}, {})
        assert res.status_code == 400
        res_json = res.json()
        expected_result = {u'status': u'error',  u'code': 400}
        for key, value in expected_result.items():
            assert res_json.get(key) == value
        if res.request.method == "GET":
            assert 'url' in res_json['message']
        else:
            assert "request" in res_json["message"]

    @pytest.mark.parametrize("method", [
        perform_get, perform_post
    ])
    def test_no_url_no_start_requests(self, server, method):
        res = method(server.url('crawl.json'), {'spider_name': 'test'},
                     {})
        assert res.status_code == 400
        expected_result = {
            u'status': u'error',
            u'code': 400
        }
        res_json = res.json()
        for key, value in expected_result.items():
            assert res_json[key] == value
        if res.request.method == "GET":
            assert 'url' in res_json['message']
        else:
            assert "request" in res_json["message"]

    @pytest.mark.parametrize("method", [
        perform_get, perform_post
    ])
    def test_no_url_but_start_requests_present(self, server, method):
        res = method(server.url("crawl.json"), {
            'spider_name': "test_with_sr",
            "start_requests": True
        }, {})
        assert res.status_code == 200
        result = res.json()
        assert result.get("status") == "ok"
        assert result.get("stats") is not None
        assert len(result.get("items", [])) == 2
        items = result["items"]
        assert len(items) == 2
        for item in items:
            name = item["name"][0]
            if name == "Page 1":
                assert "page1" in item["referer"]
            elif name == "Page 2":
                assert "page2" in item["referer"]

        spider_errors = result.get("errors", [])
        assert len(spider_errors) == 0
        assert result["stats"].get("downloader/request_count") == 2

    def test_no_request_but_start_requests_present(self, server):
        """Test for POST handler checking if everything works fine
        if there is no 'request' argument, but 'start_requests' are
        present. Not checked above because of the way default test fixtures
        are written.
        """
        post_data = {
            "no_request": {},
            "start_requests": True,
            "spider_name": "test_with_sr"
        }
        post_data.update(post_data)
        res = requests.post(server.url("crawl.json"),
                            json=post_data)
        assert res.status_code == 200
        data = res.json()
        assert len(data["items"]) == 2
        assert data.get("errors") is None

    def test_no_request_in_POST_handler(self, server):
        """Test for POST handler checking if everything works fine
        if there is no 'request' argument at all.
        """
        post_data = {
            "no_request": {},
            "spider_name": "test_with_sr"
        }
        post_data.update(post_data)
        res = requests.post(server.url("crawl.json"),
                            json=post_data)
        assert res.status_code == 400
        data = res.json()
        msg = u"Missing required parameter: 'request'"
        assert data["message"] == msg
        assert data["status"] == "error"
        assert data.get("items") is None

    @pytest.mark.parametrize("method", [
        perform_get, perform_post
    ])
    def test_url_and_start_requests_present(self, server, method):
        spider_data = {
            "url": server.target_site.url("page3.html")
        }
        api_params = {
            "spider_name": "test_with_sr",
            "start_requests": True,
        }
        res = method(server.url("crawl.json"), api_params,
                     spider_data)
        assert res.status_code == 200
        output = res.json()
        assert len(output.get("errors", [])) == 0
        items = output.get("items", [])
        assert len(items) == 3

        for item in items:
            name = item["name"][0]
            if name == "Page 1":
                assert "page1" in item["referer"]
            elif name == "Page 2":
                assert "page2" in item["referer"]
            elif name == "Page 3":
                assert item.get("referer") is None

    @pytest.mark.parametrize("method", [
        perform_get, perform_post
    ])
    def test_no_spider_name(self, server, method):
        res = method(server.url("crawl.json"),
                     {},
                     {"url": server.target_site.url("page1.html")})
        assert res.status_code == 400
        res_json = res.json()
        expected_result = {
            u'status': u'error',
            u'code': 400
        }
        for key, value in expected_result.items():
            assert res_json[key] == value
        assert 'spider_name' in res_json['message']

    def test_invalid_scrapy_request_detected_in_api(self, server):
        res = perform_post(server.url("crawl.json"),
                           {"spider_name": "test"},
                           {'url': server.target_site.url("page1.html"),
                            "not_an_argument": False})
        assert res.status_code == 400
        res_json = res.json()
        expected_result = {
            u'status': u'error',
            u'code': 400
        }
        for k, v in expected_result.items():
            assert res_json[k] == v
        assert "'not_an_argument' is not a valid arg" in res_json['message']

    @pytest.mark.parametrize("method", [
        perform_get, perform_post
    ])
    def test_invalid_scrapy_request_detected_by_scrapy(self, server, method):
        res = method(
            server.url("crawl.json"),
            {"spider_name": "test"},
            {'url': "no_rules"}
        )
        assert res.status_code == 400
        res_json = res.json()
        assert res_json["status"] == "error"
        assert res_json["code"] == 400
        assert "Error while creating Scrapy Request" in res_json['message']

    @pytest.mark.parametrize("method", [
        perform_get, perform_post
    ])
    def test_crawl(self, server, method):
        url = server.url("crawl.json")
        res = method(url,
                     {"spider_name": "test"},
                     {"url": server.target_site.url("page1.html")})
        expected_items = [{
            u'name': ['Page 1'],
        }]
        res_json = res.json()
        assert res_json["status"] == "ok"
        assert res_json["items_dropped"] == []
        assert res_json['items']
        assert len(res_json['items']) == len(expected_items)
        assert res_json["items"] == expected_items

    def test_invalid_json_in_post(self, server):
        res = requests.post(server.url("crawl.json"), data="ads")
        assert res.status_code == 400
        res_json = res.json()
        expected_result = {
            u'status': u'error',
            u'code': 400
        }
        for k, v in expected_result.items():
            assert res_json[k] == v
        msg = "Invalid JSON in POST body"
        assert msg in res_json['message']



# -*- coding: utf-8 -*-
from twisted.trial import unittest
import requests

from .servers import MockServer, ScrapyrtTestServer


class TestRootResourceIntegration(unittest.TestCase):

    def setUp(self):
        self.site = MockServer()
        self.site.start()
        self.server = ScrapyrtTestServer(site=self.site)
        self.server.start()
        self.root_url = self.server.url()

    def tearDown(self):
        if not self._passed:
            print(self.server._non_block_read(self.server.proc.stderr))
        self.server.stop()
        self.site.stop()

    def test_root(self):
        res = requests.get(self.root_url)
        assert res.status_code == 404
        assert 'No Such Resource' in res.text



# -*- coding: utf-8 -*-
import six
from subprocess import Popen, PIPE
from six.moves.urllib.parse import urljoin
import fcntl
import os
import shutil
import socket
import sys
import tempfile
import time

import port_for

from . import SAMPLE_DATA
from .utils import get_testenv, generate_project

DEVNULL = open(os.devnull, 'wb')


class BaseTestServer(object):

    def __init__(self, host='localhost', port=None, cwd=None, shell=False,
                 stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL):
        self.host = host
        self.port = port or port_for.select_random()
        self.proc = None
        self.shell = shell
        self.cwd = cwd
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        if six.PY2:
            command = 'SimpleHTTPServer'
        else:
            command = 'http.server'
        self.arguments = [
            sys.executable, '-u', '-m', command, str(self.port)
        ]

    def start(self):
        self.proc = Popen(
            self.arguments,
            stdin=self.stdin,
            stdout=self.stdout,
            stderr=self.stderr,
            shell=self.shell,
            cwd=self.cwd,
            env=get_testenv()
        )
        self.proc.poll()
        if self.proc.returncode:
            msg = (
                "unable to start server. error code: %d - stderr follows: \n%s"
            ) % (self.proc.returncode, self.proc.stderr.read())
            raise RuntimeError(msg)
        try:
            self._wait_for_port()
        finally:
            print(self._non_block_read(self.proc.stderr))
            pass

    def stop(self):
        if self.proc is None:
            raise RuntimeError("Server wasn't started")
        self.proc.kill()
        self.proc.wait()
        self.proc = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def url(self, path=''):
        return urljoin('http://{}:{}'.format(self.host, self.port), path)

    def _wait_for_port(self, delay=0.1, attempts=20):
        """Imports could take some time, server opens port with some delay."""
        while attempts > 0:
            s = socket.socket()
            try:
                s.connect((self.host, self.port))
            except Exception:
                time.sleep(delay)
                attempts -= 1
            else:
                return
            finally:
                s.close()
        raise RuntimeError("Port %d is not open" % self.port)

    @staticmethod
    def _non_block_read(output):
        if output is None:
            return ''
        fd = output.fileno()
        fl = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
        try:
            return output.read()
        except Exception:
            return ''


class ScrapyrtTestServer(BaseTestServer):

    def __init__(self, site=None, *args, **kwargs):
        super(ScrapyrtTestServer, self).__init__(*args, **kwargs)
        self.arguments = [
            sys.executable, '-m', 'scrapyrt.cmdline', '-p', str(self.port)
        ]
        self.stderr = PIPE
        self.tmp_dir = tempfile.mkdtemp()
        self.cwd = os.path.join(self.tmp_dir, 'testproject')
        generate_project(self.cwd, site=site)

    def stop(self):
        super(ScrapyrtTestServer, self).stop()
        shutil.rmtree(self.tmp_dir)


class MockServer(BaseTestServer):

    def __init__(self, *args, **kwargs):
        super(MockServer, self).__init__(*args, **kwargs)
        self.cwd = os.path.join(SAMPLE_DATA, 'testsite')



# -*- coding: utf-8 -*-
from six import StringIO

from mock import patch
from twisted.python.log import startLoggingWithObserver, removeObserver
from twisted.trial import unittest

from scrapyrt.log import ScrapyrtFileLogObserver, msg


@patch('twisted.python.log.FileLogObserver.emit')
class TestLogObserver(unittest.TestCase):

    def setUp(self):
        self.file = StringIO()
        self.log_observer = ScrapyrtFileLogObserver(self.file)
        startLoggingWithObserver(self.log_observer.emit, setStdout=False)
        self.event_dict = {'system': 'scrapyrt', 'message': 'blah'}

    def tearDown(self):
        removeObserver(self.log_observer.emit)

    def test_emit_called(self, emit_mock):
        self.log_observer.emit(self.event_dict)
        self.assertTrue(emit_mock.called)

    def test_scrapy_filtering(self, emit_mock):
        self.event_dict['system'] = 'scrapy'
        self.log_observer.emit(self.event_dict)
        self.assertFalse(emit_mock.called)

    def test_log_start_messages_filtering(self, emit_mock):
        self.event_dict['system'] = 'HTTPChannel'
        self.event_dict['message'] = 'Log opened.'
        self.log_observer.emit(self.event_dict)
        self.assertFalse(emit_mock.called)

        self.event_dict['system'] = 'other'
        self.log_observer.emit(self.event_dict)
        self.assertTrue(emit_mock.called)

    def test_unicode_message(self, emit_mock):
        original_message = u'Привет, мир!'
        msg(original_message)
        transformed_message = emit_mock.call_args[0][1]['message'][0]
        self.assertEqual(transformed_message, original_message.encode('utf-8'))



# -*- coding: utf-8 -*-
from distutils.version import LooseVersion
import pkgutil

__version__ = pkgutil.get_data(__package__, 'VERSION').decode('ascii').strip()
version_info = tuple(LooseVersion(__version__).version)

__all__ = ['__version__', 'version_info']



# -*- coding: utf-8 -*-
from scrapy.utils.engine import get_engine_status
from twisted.internet import defer
from twisted.trial import unittest

from scrapyrt.core import ScrapyrtCrawler

from .servers import MockServer
from .spiders import SingleRequestSpider
from .utils import get_settings


class TestCrawler(unittest.TestCase):
    """Spider shouldn't make start requests if list of start_requests
    wasn't passed to 'crawl' method.

    """

    def setUp(self):
        self.site = MockServer()
        self.site.start()
        self.settings = get_settings()
        self.settings['EXTENSIONS']['scrapy.extensions.corestats.CoreStats'] = 0
        self.engine_status = []

    def tearDown(self):
        self.site.stop()

    def cb(self, response):
        self.engine_status.append(get_engine_status(self.crawler.engine))

    def _assert_no_requests(self):
        self.assertEqual(len(self.engine_status), 0, self.engine_status)
        stats = self.crawler.stats.get_stats()
        self.assertNotIn('scheduler/enqueued', stats)
        self.assertNotIn('scheduler/dequeued', stats)
        self.assertNotIn('downloader/request_count', stats)
        self.assertNotIn('downloader/response_count', stats)

    def _assert_engine_worked(self):
        stats = self.crawler.stats.get_stats()
        self.assertIn('start_time', stats)
        self.assertIn('finish_time', stats)
        self.assertEquals(stats['finish_reason'], 'finished')

    @defer.inlineCallbacks
    def test_crawl_start_requests_disabled(self):
        self.crawler = ScrapyrtCrawler(
            SingleRequestSpider, self.settings, start_requests=False)
        yield self.crawler.crawl(seed=self.site.url(), callback_func=self.cb)
        self._assert_engine_worked()
        self._assert_no_requests()

    @defer.inlineCallbacks
    def test_crawl_start_requests_enabled(self):
        self.crawler = ScrapyrtCrawler(
            SingleRequestSpider, self.settings, start_requests=True)
        yield self.crawler.crawl(seed=self.site.url(), callback_func=self.cb)
        self._assert_engine_worked()
        self.assertEqual(len(self.engine_status), 1, self.engine_status)
        est = dict(self.engine_status[0])
        self.assertEqual(est['engine.spider.name'], self.crawler.spider.name)
        self.assertEqual(est['len(engine.scraper.slot.active)'], 1)
        stats = self.crawler.stats.get_stats()
        self.assertEqual(stats['scheduler/enqueued'], 1)
        self.assertEqual(stats['scheduler/dequeued'], 1)
        self.assertEqual(stats['downloader/request_count'], 1)
        self.assertEqual(stats['downloader/response_count'], 1)

    @defer.inlineCallbacks
    def test_crawl_start_requests_default(self):
        self.crawler = ScrapyrtCrawler(SingleRequestSpider, self.settings)
        yield self.crawler.crawl(seed=self.site.url(), callback_func=self.cb)
        self._assert_engine_worked()
        self._assert_no_requests()




import pytest
import sys
import tempfile
from collections import namedtuple
from os import path, chdir
from scrapy.utils.conf import closest_scrapy_cfg
from twisted.python.components import Componentized

from scrapyrt.cmdline import find_scrapy_project, get_application
from tests.utils import generate_project


def make_fake_args():
    fake_args = namedtuple('arguments', [
        'port',
        'ip',
        'set',
        'project',
        'settings'
    ])
    return fake_args(
        9080,
        '0.0.0.0',
        [],
        'default',
        ''
    )


@pytest.fixture
def workdir():
    tmp_dir = tempfile.mkdtemp()
    workdir = path.join(tmp_dir, 'testproject')
    generate_project(workdir)
    chdir(workdir)
    return workdir


class TestCmdLine(object):
    def test_find_scrapy_project(self, workdir):
        settings = find_scrapy_project('default')
        assert 'testproject.settings' == settings
        assert workdir in sys.path

    def test_find_scrapy_project_invalid_conf(self, workdir):
        config = closest_scrapy_cfg()
        with open(config, 'wb') as f:
            f.write(b'[other_section]')
        with pytest.raises(RuntimeError) as err:
            find_scrapy_project('default')
        assert str(err.value) == "No section: 'settings'"

    def test_get_application(self):
        app = get_application(make_fake_args())
        assert isinstance(app, Componentized)



# -*- coding: utf-8 -*-
import demjson
from scrapy.utils.misc import load_object
from scrapy.utils.serialize import ScrapyJSONEncoder
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure
from twisted.web import resource, server
from twisted.web.error import Error, UnsupportedMethod

from . import log
from .conf import settings
from .utils import extract_scrapy_request_args, to_bytes


# XXX super() calls won't work wihout object mixin in Python 2
# maybe this can be removed at some point?
class ServiceResource(resource.Resource, object):
    json_encoder = ScrapyJSONEncoder()

    def __init__(self, root=None):
        resource.Resource.__init__(self)
        self.root = root

    def render(self, request):
        try:
            result = resource.Resource.render(self, request)
        except Exception as e:
            result = self.handle_error(e, request)

        if not isinstance(result, Deferred):
            return self.render_object(result, request)

        # deferred result - add appropriate callbacks and errbacks
        result.addErrback(self.handle_error, request)

        def finish_request(obj):
            request.write(self.render_object(obj, request))
            request.finish()

        result.addCallback(finish_request)
        return server.NOT_DONE_YET

    def handle_error(self, exception_or_failure, request):
        """Override this method to add custom exception handling.

        :param request: twisted.web.server.Request
        :param exception_or_failure: Exception or
            twisted.python.failure.Failure
        :return: dict which will be converted to JSON error response

        """
        failure = None
        if isinstance(exception_or_failure, Exception):
            exception = exception_or_failure
        elif isinstance(exception_or_failure, Failure):
            exception = exception_or_failure.value
            failure = exception_or_failure
        else:
            raise TypeError(
                'Expected Exception or {} instances, got {}'.format(
                    Failure,
                    exception_or_failure.__class__
                ))
        if request.code == 200:
            # Default code - means that error wasn't handled
            if isinstance(exception, UnsupportedMethod):
                request.setResponseCode(405)
            elif isinstance(exception, Error):
                code = int(exception.status)
                request.setResponseCode(code)
            else:
                request.setResponseCode(500)
            if request.code == 500:
                log.err(failure)
        return self.format_error_response(exception, request)

    def format_error_response(self, exception, request):
        # Python exceptions don't have message attribute in Python 3+ anymore.
        # Twisted HTTP Error objects still have 'message' attribute even in 3+
        # and they fail on str(exception) call.
        msg = exception.message if hasattr(exception, 'message') else str(exception)
        return {
            "status": "error",
            "message": msg,
            "code": request.code
        }

    def render_object(self, obj, request):
        r = self.json_encoder.encode(obj) + "\n"
        request.setHeader('Content-Type', 'application/json')
        request.setHeader('Access-Control-Allow-Origin', '*')
        request.setHeader('Access-Control-Allow-Methods',
                          ', '.join(getattr(self, 'allowedMethods', [])))
        request.setHeader('Access-Control-Allow-Headers', 'X-Requested-With')
        request.setHeader('Content-Length', str(len(r)))
        return r.encode("utf8")


class RealtimeApi(ServiceResource):

    def __init__(self, **kwargs):
        super(RealtimeApi, self).__init__(self)
        for route, resource_path in settings.RESOURCES.items():
            resource_cls = load_object(resource_path)
            route = to_bytes(route)
            self.putChild(route, resource_cls(self, **kwargs))


class CrawlResource(ServiceResource):

    isLeaf = True
    allowedMethods = ['GET', 'POST']

    def render_GET(self, request, **kwargs):
        """Request querysting must contain following keys: url, spider_name.

        At the moment kwargs for scrapy request are not supported in GET.
        They are supported in POST handler.
        """
        api_params = dict(
            (name.decode('utf-8'), value[0].decode('utf-8'))
            for name, value in request.args.items()
        )
        scrapy_request_args = extract_scrapy_request_args(api_params,
                                                          raise_error=False)
        self.validate_options(scrapy_request_args, api_params)
        return self.prepare_crawl(api_params, scrapy_request_args, **kwargs)

    def render_POST(self, request, **kwargs):
        """
        :param request:
            body should contain JSON

        Required keys in JSON posted:

        :spider_name: string
            name of spider to be scheduled.

        :request: json object
            request to be scheduled with spider.
            Note: request must contain url for spider.
            It may contain kwargs to scrapy request.

        """
        request_body = request.content.getvalue()
        try:
            api_params = demjson.decode(request_body)
        except demjson.JSONDecodeError as e:
            message = "Invalid JSON in POST body. {}"
            message = message.format(e.pretty_description())
            raise Error('400', message=message)

        log.msg("{}".format(api_params))
        if api_params.get("start_requests"):
            # start requests passed so 'request' argument is optional
            _request = api_params.get("request", {})
        else:
            # no start_requests, 'request' is required
            _request = self.get_required_argument(api_params, "request")
        try:
            scrapy_request_args = extract_scrapy_request_args(
                _request, raise_error=True
            )
        except ValueError as e:
            raise Error('400', str(e))

        self.validate_options(scrapy_request_args, api_params)
        return self.prepare_crawl(api_params, scrapy_request_args, **kwargs)

    def validate_options(self, scrapy_request_args, api_params):
        url = scrapy_request_args.get("url")
        start_requests = api_params.get("start_requests")
        if not url and not start_requests:
            raise Error('400',
                        "'url' is required if start_requests are disabled")

    def get_required_argument(self, api_params, name, error_msg=None):
        """Get required API key from dict-like object.

        :param dict api_params:
            dictionary with names and values of parameters supplied to API.
        :param str name:
            required key that must be found in api_params
        :return: value of required param
        :raises Error: Bad Request response

        """
        if error_msg is None:
            error_msg = 'Missing required parameter: {}'.format(repr(name))
        try:
            value = api_params[name]
        except KeyError:
            raise Error('400', message=error_msg)
        if not value:
            raise Error('400', message=error_msg)
        return value

    def prepare_crawl(self, api_params, scrapy_request_args, *args, **kwargs):
        """Schedule given spider with CrawlManager.

        :param dict api_params:
            arguments needed to find spider and set proper api parameters
            for crawl (max_requests for example)

        :param dict scrapy_request_args:
            should contain positional and keyword arguments for Scrapy
            Request object that will be created
        """
        spider_name = self.get_required_argument(api_params, 'spider_name')
        start_requests = api_params.get("start_requests", False)
        try:
            max_requests = api_params['max_requests']
        except (KeyError, IndexError):
            max_requests = None
        dfd = self.run_crawl(
            spider_name, scrapy_request_args, max_requests,
            start_requests=start_requests, *args, **kwargs)
        dfd.addCallback(
            self.prepare_response, request_data=api_params, *args, **kwargs)
        return dfd

    def run_crawl(self, spider_name, scrapy_request_args,
                  max_requests=None, start_requests=False, *args, **kwargs):
        crawl_manager_cls = load_object(settings.CRAWL_MANAGER)
        manager = crawl_manager_cls(spider_name, scrapy_request_args, max_requests, start_requests=start_requests)
        dfd = manager.crawl(*args, **kwargs)
        return dfd

    def prepare_response(self, result, *args, **kwargs):
        items = result.get("items")
        response = {
            "status": "ok",
            "items": items,
            "items_dropped": result.get("items_dropped", []),
            "stats": result.get("stats"),
            "spider_name": result.get("spider_name"),
        }
        errors = result.get("errors")
        if errors:
            response["errors"] = errors
        return response



# -*- coding: utf-8 -*-
from scrapy import Spider, Request


class MetaSpider(Spider):
    """Copy-paste from scrapy tests."""

    name = 'meta'

    def __init__(self, *args, **kwargs):
        super(MetaSpider, self).__init__(*args, **kwargs)
        self.meta = {}

    def closed(self, reason):
        self.meta['close_reason'] = reason


class SingleRequestSpider(MetaSpider):
    """Copy-paste from scrapy tests."""
    seed = None
    callback_func = None
    errback_func = None
    name = 'single_request'

    def start_requests(self):
        if isinstance(self.seed, Request):
            yield self.seed.replace(callback=self.parse, errback=self.on_error)
        else:
            yield Request(self.seed, callback=self.parse, errback=self.on_error)

    def parse(self, response):
        self.meta.setdefault('responses', []).append(response)
        if callable(self.callback_func):
            return self.callback_func(response)
        if 'next' in response.meta:
            return response.meta['next']

    def on_error(self, failure):
        self.meta['failure'] = failure
        if callable(self.errback_func):
            return self.errback_func(failure)



# -*- coding: utf-8 -*-
#!/usr/bin/python
from setuptools import setup, find_packages
from os.path import join, dirname

with open(join(dirname(__file__), 'scrapyrt/VERSION'), 'rb') as f:
    version = f.read().decode('ascii').strip()

setup(
    name="scrapyrt",
    version=version,
    author='Scrapinghub',
    author_email='info@scrapinghub.com',
    url="https://github.com/scrapinghub/scrapyrt",
    maintainer='Scrapinghub',
    maintainer_email='info@scrapinghub.com',
    description='Put Scrapy spiders behind an HTTP API',
    long_description=open('README.rst').read(),
    license='BSD',
    packages=find_packages(),
    entry_points={
        'console_scripts': ['scrapyrt = scrapyrt.cmdline:execute']
    },
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent',
        'Environment :: Console',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Internet :: WWW/HTTP',
        'License :: OSI Approved :: BSD License',
    ],
    install_requires=[
        'Twisted>=14.0.0',
        'Scrapy>=1.0.0',
        'demjson',
        'six>=1.5.2'
    ],
    package_data={
        'scrapyrt': [
            'VERSION',
        ]
    },
)



# -*- coding: utf-8 -*-
import six
from copy import deepcopy
from importlib import import_module

from . import default_settings


class Settings(object):

    def __init__(self):
        self.setmodule(default_settings)

    def setmodule(self, module):
        if isinstance(module, six.string_types):
            module = import_module(module)
        for setting in dir(module):
            self.set(setting, getattr(module, setting))

    def __setattr__(self, key, value):
        if self.frozen:
            raise TypeError("Trying to modify a frozen Settings object")
        return super(Settings, self).__setattr__(key, value)

    def set(self, name, value):
        if not name.startswith('_') and name.isupper():
            # Deepcopy objects here, or we will have issues with mutability,
            # like changing mutable object stored in settings leads to
            # change of object in default_settings module.
            setattr(self, name, deepcopy(value))

    def freeze(self):
        self._frozen = True

    @property
    def frozen(self):
        return bool(getattr(self, '_frozen', False))


settings = Settings()



# -*- coding: utf-8 -*-
from functools import wraps
import warnings

from .exceptions import ScrapyrtDeprecationWarning


def deprecated(use_instead=None):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.

    Taken from Scrapy.
    """

    def deco(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            message = "Call to deprecated function %s." % func.__name__
            if use_instead:
                message += " Use %s instead." % use_instead
            warnings.warn(message, category=ScrapyrtDeprecationWarning,
                          stacklevel=2)
            return func(*args, **kwargs)

        return wrapped

    if callable(use_instead):
        deco = deco(use_instead)
        use_instead = None
    return deco



# -*- coding: utf-8 -*-
from six.moves.configparser import (
    SafeConfigParser, NoOptionError, NoSectionError
)
import argparse
import os
import sys

from scrapy.utils.conf import closest_scrapy_cfg
from scrapy.utils.misc import load_object
from twisted.application import app
from twisted.application.internet import TCPServer
from twisted.application.service import Application
from twisted.internet import reactor
from twisted.web.server import Site

from .log import setup_logging
from .conf import settings


def parse_arguments():

    def valid_setting(string):
        key, sep, value = string.partition('=')
        if not key or not sep:
            raise argparse.ArgumentTypeError(
                u'expected name=value: {}'.format(repr(string)))
        return key, value

    parser = argparse.ArgumentParser(
        description='HTTP API server for Scrapy project.')
    parser.add_argument('-p', '--port', dest='port',
                        type=int,
                        default=9080,
                        help='port number to listen on')
    parser.add_argument('-i', '--ip', dest='ip',
                        default='localhost',
                        help='IP address the server will listen on')
    parser.add_argument('--project', dest='project',
                        default='default',
                        help='project name from scrapy.cfg')
    parser.add_argument('-s', '--set', dest='set',
                        type=valid_setting,
                        action='append',
                        default=[],
                        metavar='name=value',
                        help='set/override setting (may be repeated)')
    parser.add_argument('-S', '--settings', dest='settings',
                        metavar='project.settings',
                        help='custom project settings module path')
    return parser.parse_args()


def get_application(arguments):
    ServiceRoot = load_object(settings.SERVICE_ROOT)
    site = Site(ServiceRoot())
    application = Application('scrapyrt')
    server = TCPServer(arguments.port, site, interface=arguments.ip)
    server.setServiceParent(application)
    return application


def find_scrapy_project(project):
    project_config_path = closest_scrapy_cfg()
    if not project_config_path:
        raise RuntimeError('Cannot find scrapy.cfg file')
    project_config = SafeConfigParser()
    project_config.read(project_config_path)
    try:
        project_settings = project_config.get('settings', project)
    except (NoSectionError, NoOptionError) as e:
        raise RuntimeError(e.message)
    if not project_settings:
        raise RuntimeError('Cannot find scrapy project settings')
    project_location = os.path.dirname(project_config_path)
    sys.path.append(project_location)
    return project_settings


def execute():
    sys.path.insert(0, os.getcwd())
    arguments = parse_arguments()
    if arguments.settings:
        settings.setmodule(arguments.settings)
    if arguments.set:
        for name, value in arguments.set:
            settings.set(name.upper(), value)
    settings.set('PROJECT_SETTINGS', find_scrapy_project(arguments.project))
    settings.freeze()
    setup_logging()
    application = get_application(arguments)
    app.startApplication(application, save=False)
    reactor.run()


if __name__ == '__main__':
    execute()



# -*- coding: utf-8 -*-
from copy import deepcopy
import os

from mock import patch
from twisted.trial import unittest

from scrapyrt.conf import settings
from scrapyrt.resources import RealtimeApi, ServiceResource, CrawlResource


class TestResource(ServiceResource):
    isLeaf = True
    allowedMethods = ['GET', 'POST']

    def render_GET(self, request, **kwargs):
        return {'status': 'ok'}


class TestRealtimeApi(unittest.TestCase):

    @staticmethod
    def _get_class_path(clsname):
        module_name, _, _ = os.path.basename(__file__).rpartition('.')
        return '{}.{}.{}'.format(__package__, module_name, clsname)

    def test_realtimeapi_with_default_settings(self):
        expected_entities = {b'crawl.json': CrawlResource}
        service_root = RealtimeApi()
        self._check_entities(service_root, expected_entities)

    # XXX: one inconvenience of singleton settings - complexities during tests,
    # e.g. settings are mutable, when you change them in one test -
    # changes will be kept unless you cleanup those changes or use mock.
    @patch('scrapyrt.resources.settings', deepcopy(settings))
    def test_realtimeapi_with_custom_settings(self):
        from scrapyrt.resources import settings
        settings.RESOURCES[b'test.json'] = self._get_class_path('TestResource')
        expected_entities = {
            b'crawl.json': CrawlResource,
            b'test.json': TestResource
        }
        service_root = RealtimeApi()
        self._check_entities(service_root, expected_entities)

    def _check_entities(self, service_root, expected_entities):
        self.assertFalse(service_root.isLeaf)
        entities = service_root.listEntities()
        self.assertEqual(len(entities), len(expected_entities))
        for name, entity in entities:
            self.assertIn(name, expected_entities)
            self.assertIsInstance(entity, expected_entities[name])
